/* ************************************************************************ 
 *         The Amulet User Interface Development Environment              *
 * ************************************************************************
 * This code was written as part of the Amulet project at                 *
 * Carnegie Mellon University, and has been placed in the public          *
 * domain.  If you are using this code or any part of Amulet,             *
 * please contact amulet@cs.cmu.edu to be put on the mailing list.        *
 * ************************************************************************/

#include <am_inc.h>

//#define USE_SOCKETS

#ifdef USE_SOCKETS
#include "..\pebbles\PebblesSocketMessages.h"
#else
#include "pebblesmulticursor.h" // internal pebbles interface
#endif

#include "MultiCursorMessages.h" // codes for all the messages
#include PEBBLES_AMULET__H   // Am_Pebbles_Init, Am_Handle_Pebbles_Event
#include IDEFS__H
#include GEM__H
#include OPAL_ADVANCED__H
#include OBJECT__H

#include "gemW.h" // need some of this junk to import gemW_draw.h
#include "gemW_misc.h"   // just to get the Am_WinDrawonable
#include "gemW_clean.h"
#include "gemW_points.h"
#include "gemW_styles.h"
#include "gemW_text.h"
#include "gemW_line.h"
#include "gemW_input.h"
#include "gemW_regions.h"
#include <fstream.h>  // for images
#include "gemW_image.h"
#include "gemW_cursor.h"

#include "gemW_draw.h"  // Am_WinDrawonable

// This procedure is registered by Am_Pebbles_Init and called by
// gemW_pebbles for each new user, including those that exist when
// init is called.
Am_Pebbles_New_User_Proc* global_new_user_proc;

Am_Handle_Pebbles_Socket_Event_Proc_Type* am_pebbles_socket_handler;

Am_Input_Event pebbles_ev;

class User_Struct {
public:
	int user_id;
	int cur_x;
	int cur_y;
	int last_x;
	int last_y;
	char* name;
	bool shift_key_down;
	bool caps_lock_key_down;
	bool control_key_down;
	bool alt_key_down;
	bool drag;

	User_Struct()
		{ user_id = -1; // will be my index + 1 (so not 0)
		  cur_x = 0;
		  cur_y = 0;
		  last_x = 0;
		  last_y = 0;
		  name = NULL;
		  shift_key_down = false;
		  caps_lock_key_down = false;
		  control_key_down = false;
		  alt_key_down = false;
		  drag = false; }
};

static bool initialized = false;
User_Struct local_users[MAX_MULTICURSOR_USERS];

#define DLL_NAME     "MultiCursor"
#define ENTRY_POINT  "Init_Pebbles_MultiCursor"
#define PEBBLES_USER_NAME   "Get_Pebbles_User_Name"

typedef  void  (CALLBACK*EntryPoint) (HWND);
typedef  char*  (CALLBACK*GetPebblesUserNameType) (int);

GetPebblesUserNameType GetPebblesUserName;

void handle_new_user(int user_index, char * name) {
	local_users[user_index].name = new char[strlen(name)+1];
	strcpy(local_users[user_index].name, name);
	local_users[user_index].user_id = user_index;

	cout << "New user, id " << local_users[user_index].user_id
			<< " name " << local_users[user_index].name << endl << flush;
	global_new_user_proc(local_users[user_index].user_id,
		local_users[user_index].name);
}

void find_dll_and_set_pebbles (HWND hwnd) {
    HINSTANCE hDLL = LoadLibrary (DLL_NAME);
    if (!hDLL) 
		Am_ERROR("Can't find dll library " << DLL_NAME);

    EntryPoint proc = (EntryPoint)GetProcAddress(hDLL, ENTRY_POINT);
    if (!proc) 
		Am_ERROR("Can't find procedure " << ENTRY_POINT
				 << " in dll library " << DLL_NAME);
    proc (hwnd);

    GetPebblesUserName = (GetPebblesUserNameType)GetProcAddress(hDLL, PEBBLES_USER_NAME);
    if (!GetPebblesUserName)
		Am_ERROR("Can't find procedure " << PEBBLES_USER_NAME
				 << " in dll library " << DLL_NAME);

	// copy all the initial users
	char * name;
	int user_index = 0;
    for (user_index = 0; user_index < MAX_MULTICURSOR_USERS; user_index++) {
		local_users[user_index].user_id = user_index+1;
		name = GetPebblesUserName(user_index);
		if (name) {
			if (local_users[user_index].name) {
				// user has name already, so forget it
				}
			else handle_new_user(user_index, name);
		}
	}

    // FreeLibrary (hDLL); need the library connected so can get the users
}

// Initialize the Pebbles MultiCursor software with the main window
// associated with this program.  This cannot be called until the
// hWnd is fully created.
void Am_Pebbles_Init(Am_Object & window, Am_Pebbles_New_User_Proc proc) {
	cout << "()() Calling Init_Pebbles\n" << flush;
	global_new_user_proc = proc;
	
#ifndef USE_SOCKETS
	Am_Drawonable* draw = Get_a_drawonable(window);
	if (draw) {
		HWND hWnd = ((Am_WinDrawonable*)draw)->WndGet();
		find_dll_and_set_pebbles(hWnd);
		cout << "found dll (not sockets)\n" << flush;
		}
	else Am_ERROR("No drawonable for Pebbles on window " << window);
#endif
	initialized = true;
}


User_Struct* find_user_struct(int want_user_id) {
	if (local_users[want_user_id].name)
		return &(local_users[want_user_id]);
	else {
		cout << "user for id " << want_user_id
				 << " not initialized" << endl; //shouldn't happen

		return NULL; //avoid warnings
	}
}

static void i_process_event(Am_Input_Char ic, int x, int y, int user_id,
									Am_Drawonable* drw) {
	pebbles_ev.Set(ic, x, y, 0, drw, user_id);
	Am_WinDrawonable* win_drw = (Am_WinDrawonable*)drw;
	win_drw -> m_evh -> Input_Event_Notify(drw, &pebbles_ev);
    if (Am_Debug_Print_Input_Events) {
      Am_TRACE("<P><P> " << ev -> input_char <<
	    " t = " << ev -> time_stamp  <<
	  	" user_id=" << ev -> user_id  <<
		endl);
	}
} 
static void i_process_event(Am_Input_Char ic, User_Struct* user,
									Am_Drawonable* drw) {
	i_process_event(ic, user->cur_x, user->cur_y, user->user_id, drw);
}

static void i_process_regular_key(short key, User_Struct* user,
								  Am_Drawonable* drw) {
	bool shift = user->shift_key_down;
	if (key >= 'A' && key <= 'Z') shift = true;
	else if (user->caps_lock_key_down && key >= 'a' && key <= 'z')
		shift = true;
	if (shift && key >= 'a' && key <= 'z')
		key = key - 'a' + 'A'; // make sure key and shift are consistent
	Am_Input_Char ic(key, shift,
		user->control_key_down, user->alt_key_down,
		Am_NEITHER, Am_NOT_MOUSE);
	i_process_event(ic, user, drw);
}

static void i_process_mouse_event(long code, Am_Button_Down down, 
									int x, int y, User_Struct* user,
									Am_Drawonable* drw) {
	Am_Input_Char ic(code, user->shift_key_down,
		user->control_key_down, user->alt_key_down,
		down, Am_SINGLE_CLICK);
	i_process_event(ic, x, y, user->user_id, drw);
}

static void i_process_mouse_event(long code, Am_Button_Down down, User_Struct* user,
									Am_Drawonable* drw) {
	i_process_mouse_event(code, down, user->cur_x, user->cur_y, user, drw);
}

inline unsigned char get3(LPARAM l) {
	return (l & 0x0000FF00) >> 8;
}
inline unsigned char get4(LPARAM l) {
	return (l & 0x00FF0000) >> 16;
}

//this should be called for all mouse up and move events
//it modifies x and y of User_Struct and returns the new drawonable if necessary.  If not
// necessary, just returns drw
static Am_Drawonable* check_multi_window(Am_Drawonable* drw, HWND hWnd,
							int inx, int iny, int &outx, int &outy,
							bool is_move_event) {
	//check for multiple windows movement
	// ** this relatively expensive check has to be done on every event !!
	//  ** How make this faster?
	Am_WinDrawonable* win_drw = (Am_WinDrawonable*)drw;
	if (win_drw->m_want_multi_window) {
		POINT pt;
		pt.x = inx;
		pt.y = iny;
		ClientToScreen(hWnd, &pt);
		HWND hWndM = WindowFromPoint(pt);
		Am_WinDrawonable* drwM;
		// If it's a different window,
		// and the window exists,
		// and it's associated with an Amulet drawonable, then
		// convert the event location to new window coordinates. 
		// (don't check multi-window since for pebbles, it isn't necessarily
		//  dragging anything)
		if (hWndM != hWnd && hWndM && 
			(drwM = Am_WinDrawonable::Narrow(Am_WinDrawonable::m_mapWnd2Drw.GetAt(hWndM)))) {
			ScreenToClient(hWndM, &pt);
			outx = pt.x;
			outy = pt.y;
			return drwM;
			}
		}
	outx = inx;
	outy = iny;
	return drw;
}

void check_accel_inc(int &xinc, int &yinc) {
	int amt = abs(xinc);
	if (abs(yinc) > amt) amt = abs(yinc);
	if (amt <= 2) ; // nothing
	else if (amt <= 4) { xinc *= 2; yinc *= 2; }
	else if (amt <= 6) { xinc *= 3; yinc *= 3; }
	else			   { xinc *= 4; yinc *= 4; }
}

void Am_Handle_Pebbles_Socket_Event(Am_Object & window, WPARAM wParam, LPARAM lParam) {
	Am_Drawonable* draw = Get_a_drawonable(window);
	if (draw) {
		HWND hWnd = ((Am_WinDrawonable*)draw)->WndGet();
		am_pebbles_handler(hWnd, draw, wParam, lParam);
		}
	else Am_ERROR("No drawonable for Pebbles on window " << window);
	
}

void Am_Pebbles_Set_User_Cursor_Pos(int user_id,
									int new_x, int new_y) {
	User_Struct* user = find_user_struct(user_id);
	user->cur_x = new_x;
	user->cur_y = new_y;
}

// may change the drawonable, if not, just returns drw
void Am_Handle_Pebbles_Event(HWND hWnd, Am_Drawonable* drw, 
							 WPARAM wParam, LPARAM lParam) {

	if (!initialized) Am_Error("Pebbles event but not initialized");

	int user_id_int = wParam;

    unsigned char message_type = lParam & 0x000000FF; // lsb is command type

    char chr;
    unsigned char x,y;
	int outx, outy;
    short result;
    char *temp;
    char modifiers;
    char vkey;
	User_Struct* user = NULL;

	user = find_user_struct(user_id_int);

#ifdef USE_SOCKETS
	if(!user) {
		// CMD_NEW_USER message

		//char * name = GetPebblesUserName(user_id_int);
		char *name = (char *) lParam;
		cout << "New user " << user_id_int << " name: " << name << endl << flush;
		if (local_users[user_id_int].name) {
			cout << "  but already have name " 
					<< local_users[user_id_int].name << endl << flush;
		}
		else handle_new_user(user_id_int, name);

	}
#endif


  switch(message_type) {

#ifndef USE_SOCKETS
	case CMD_NEW_USER: {
		char * name = GetPebblesUserName(user_id_int);
		cout << "New user " << user_id_int << endl << flush;
		if (name) {
			if (local_users[user_id_int].name) {
				cout << "  but already have name " 
						<< local_users[user_id_int].name << endl << flush;
			}
			else handle_new_user(user_id_int, name);
		}
		else { // name == NULL
			cout << " user " << user_id_int << " leaving\n" << flush;
			// ignore that they are leaving
		}
	break;
	}
#else
	case CMD_DONE_USER: {
		cout << " user " << user_id_int << " leaving\n" << flush;
		break;
	}	
#endif
	case CMD_MOVE: {
		x = get3(lParam);
		y = get4(lParam);
		int xinc = (int)(char)x; // (x - user->last_x);
		int yinc = (int)(char)y; // (y - user->last_y);
		if (xinc != 0 || yinc != 0) { // otherwise, don't do any further processing
			check_accel_inc(xinc, yinc);
			user->cur_x += xinc;
			user->cur_y += yinc;
			user->last_x = x;
			user->last_y = y;
			if (user->cur_x < 0) user->cur_x = 0; //peg at 0
			if (user->cur_y < 0) user->cur_y = 0; //peg at 0
			Am_Input_Char ic(Am_MOUSE_MOVED, user->shift_key_down,
				user->control_key_down, user->alt_key_down,
				Am_NEITHER, Am_SINGLE_CLICK);
			// check if there is another pending move event
			MSG peekmsg;
			 if (PeekMessage(&peekmsg, hWnd, PEBBLESMULTICURSORMESSAGE,
								PEBBLESMULTICURSORMESSAGE, PM_NOREMOVE))
				if (peekmsg.message == PEBBLESMULTICURSORMESSAGE
				    && (peekmsg.lParam & 0x000000FF) == CMD_MOVE)
					break; // don't bother processing this one
			drw = check_multi_window(drw, hWnd, user->cur_x, user->cur_y, outx, outy, true);
			i_process_event(ic, outx, outy, user->user_id, drw);
			break;
		} // if
		} //case
	case CMD_PEN_DOWN:
		user->last_x = get3(lParam); //relative motion
		user->last_y = get4(lParam);
		break;
	case CMD_PEN_UP:
		break;
	case CMD_LEFTDOWN:
		if (user->drag) {
			user->drag = false;
			drw = check_multi_window(drw, hWnd, user->cur_x, user->cur_y, outx, outy, false);
			i_process_mouse_event(Am_LEFT_MOUSE, Am_BUTTON_UP, outx, outy, user, drw);
		}
		else {
			user->drag = true;
			drw = check_multi_window(drw, hWnd, user->cur_x, user->cur_y, outx, outy, false);
			i_process_mouse_event(Am_LEFT_MOUSE, Am_BUTTON_DOWN, outx, outy, user, drw);
		}
		break;
	case CMD_LEFTUP:
		user->drag = false;
		drw = check_multi_window(drw, hWnd, user->cur_x, user->cur_y, outx, outy, false);
		i_process_mouse_event(Am_LEFT_MOUSE, Am_BUTTON_UP, outx, outy, user, drw);
		break;
	case CMD_RIGHTDOWN:
		drw = check_multi_window(drw, hWnd, user->cur_x, user->cur_y, outx, outy, false);
		i_process_mouse_event(Am_RIGHT_MOUSE, Am_BUTTON_DOWN, outx, outy, user, drw);
		break;
	case CMD_RIGHTUP:
		drw = check_multi_window(drw, hWnd, user->cur_x, user->cur_y, outx, outy, false);
		i_process_mouse_event(Am_RIGHT_MOUSE, Am_BUTTON_UP, outx, outy, user, drw);
		break;
	case CMD_KEYDOWN: {
		char chr = get3(lParam);
		i_process_regular_key(chr, user, drw);
		break;
		}
	case CMD_CONTROLDOWN:
		user->control_key_down = true;
		break;
	case CMD_CONTROLUP:
		user->control_key_down = false;
		break;
	case CMD_ALTDOWN:
		user->alt_key_down = true;
		break;
	case CMD_ALTUP:
		user->alt_key_down = false;
		break;
	case CMD_SHIFTDOWN:
		user->shift_key_down = true;
		break;
	case CMD_SHIFTUP:
		user->shift_key_down = false;
		break;
	case CMD_CAPHIT:
		user->caps_lock_key_down = !user->caps_lock_key_down;
		break;
	case CMD_ESCAPEHIT:
		i_process_regular_key(Am_ESC, user, drw);
		break;
	case CMD_INSERTHIT:
		i_process_regular_key(Am_INSERT, user, drw);
		break;
	case CMD_DELETEHIT:
		i_process_regular_key(Am_DELETE, user, drw);
		break;
	case CMD_F1HIT:
		i_process_regular_key(Am_F1, user, drw);
		break;
	case CMD_F2HIT:
		i_process_regular_key(Am_F2, user, drw);
		break;
	case CMD_F3HIT:
		i_process_regular_key(Am_F3, user, drw);
		break;
	case CMD_F4HIT:
		i_process_regular_key(Am_F4, user, drw);
		break;
	case CMD_F5HIT:
		i_process_regular_key(Am_F5, user, drw);
		break;
	case CMD_F6HIT:
		i_process_regular_key(Am_F6, user, drw);
		break;
	case CMD_F7HIT:
		i_process_regular_key(Am_F7, user, drw);
		break;
	case CMD_F8HIT:
		i_process_regular_key(Am_F8, user, drw);
		break;
	case CMD_F9HIT:
		i_process_regular_key(Am_F9, user, drw);
		break;
	case CMD_F10HIT:
		i_process_regular_key(Am_F10, user, drw);
		break;
	case CMD_F11HIT:
		i_process_regular_key(Am_F11, user, drw);
		break;
	case CMD_F12HIT:
		i_process_regular_key(Am_F12, user, drw);
		break;
	case CMD_HOMEHIT:
		i_process_regular_key(Am_HOME, user, drw);
		break;
	case CMD_ENDHIT:
		i_process_regular_key(Am_END, user, drw);
		break;
	case CMD_PAGEUPHIT:
		i_process_regular_key(Am_PAGE_UP, user, drw);
		break;
	case CMD_PAGEDOWNHIT:
		i_process_regular_key(Am_PAGE_DOWN, user, drw);
		break;
	case CMD_LEFTHIT:
		i_process_regular_key(Am_LEFT_ARROW, user, drw);
		break;
	case CMD_RIGHTHIT:
		i_process_regular_key(Am_RIGHT_ARROW, user, drw);
		break;
	case CMD_UPHIT:
		i_process_regular_key(Am_UP_ARROW, user, drw);
		break;
	case CMD_DOWNHIT:
		i_process_regular_key(Am_DOWN_ARROW, user, drw);
		break;
	case CMD_PRINTSCREENHIT:
		i_process_regular_key(Am_PRINT_KEY, user, drw);
		break;
	case CMD_PAUSEHIT:
		i_process_regular_key(Am_BREAK, user, drw);  // can't find Am_PAUSE
		break;
	case CMD_BREAKHIT:
		i_process_regular_key(Am_BREAK, user, drw);
		break;
	default:
		char* temp = new char[80];
		sprintf(temp,"What's this?! (RemoteControl) %d!\n",message_type);
		OutputDebugString (temp);
		delete [] temp;
		break;
    }
}

void Am_Amulet_Pebbles_Initialize() {

	am_pebbles_handler = &Am_Handle_Pebbles_Event;
	am_pebbles_socket_handler = &Am_Handle_Pebbles_Socket_Event;
}

#include INITIALIZER__H

Am_Initializer* pebbles_amulet_init =
	new Am_Initializer("Pebbles", Am_Amulet_Pebbles_Initialize, 1.9);
